import argparse
import time
import unittest

import dbinfo
import numpy  as np

from aperturedb import Connector, Status
from aperturedb import EntityLoader, ConnectionLoader
from aperturedb import ImageLoader, BBoxLoader, BlobLoader
from aperturedb import DescriptorSetLoader, DescriptorLoader

class TestLoader(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # ApertureDB Server Info
        self.db_host = dbinfo.DB_HOST
        self.db_port = dbinfo.DB_PORT

        # Config params
        self.batchsize  = 99
        self.numthreads = 31
        self.stats      = False

        db_up = False
        attempts = 0
        while(not db_up):
            try:
                db = Connector.Connector(self.db_host, self.db_port)
                db_up = True
                if (attempts > 0):
                    print("Connection to ApertureDB successful.")
            except:
                print("Attempt", attempts,
                      "to connect to ApertureDB failed, retying...")
                attempts += 1
                time.sleep(1) # sleeps 1 second

            if attempts > 10:
                print("Failed to connect to ApertureDB after 10 attempts")
                exit()

class TestEntityLoader(TestLoader):

    def test_Loader(self):

        # Insert Person Nodes
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/persons.adb.csv"

        generator = EntityLoader.EntityGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = EntityLoader.EntityLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_entities("Person"))


        # Insert Images
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/images.adb.csv"

        generator = ImageLoader.ImageGeneratorCSV(in_csv_file, check_image=False)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_images())


        # Insert Connections
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/connections-persons-images.adb.csv"

        generator = ConnectionLoader.ConnectionGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = ConnectionLoader.ConnectionLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_connections("has_image"))

        # Insert BBoxes
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/bboxes.adb.csv"

        generator = BBoxLoader.BBoxGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = BBoxLoader.BBoxLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_bboxes())

    def test_LoaderDescriptorsImages(self):

        # Insert Images, images may be there already, and that case should
        # be handled correctly by contraints
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/images.adb.csv"

        generator = ImageLoader.ImageGeneratorCSV(in_csv_file, check_image=False)

        if self.stats:
            print("\n")

        loader = ImageLoader.ImageLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_images())


        # Insert DescriptorsSet
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/descriptorset.adb.csv"

        generator = DescriptorSetLoader.DescriptorSetGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = DescriptorSetLoader.DescriptorSetLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_entities("_DescriptorSet"))

        # Insert Descriptors

        sets = ["setA", "setB"]

        total_descriptors = 0
        for setname in sets:
            db = Connector.Connector(self.db_host, self.db_port)

            in_csv_file = "./input/" + setname + ".adb.csv"

            generator = DescriptorLoader.DescriptorGeneratorCSV(in_csv_file)

            if self.stats:
                print("\n")

            loader = DescriptorLoader.DescriptorLoader(db)
            loader.ingest(generator, batchsize=self.batchsize,
                                     numthreads=self.numthreads,
                                     stats=self.stats)

            dbstatus = Status.Status(db)
            total_descriptors += len(generator)
            self.assertEqual(total_descriptors, dbstatus.count_entities("_Descriptor"))

    def test_BlobLoader(self):

        # Insert Person Nodes
        db = Connector.Connector(self.db_host, self.db_port)

        in_csv_file = "./input/blobs.adb.csv"

        generator = BlobLoader.BlobGeneratorCSV(in_csv_file)

        if self.stats:
            print("\n")

        loader = BlobLoader.BlobLoader(db)
        loader.ingest(generator, batchsize=self.batchsize,
                                 numthreads=self.numthreads,
                                 stats=self.stats)

        dbstatus = Status.Status(db)
        self.assertEqual(len(generator), dbstatus.count_entities("_Blob"))

        db = Connector.Connector(self.db_host, self.db_port)

        query = [{
            "FindBlob": {
                "blobs": True,
                "results": {
                    "list": ["license"],
                    "limit": 1
                }
            }
        }]

        res, blob = db.query(query)
        self.assertEqual(len(blob), 1)

        arr = np.frombuffer(blob[0])
        self.assertEqual(arr[2], 3.3)



import math
import time
from threading import Thread

from aperturedb import ProgressBar

import numpy as np

class ParallelLoader:

    """
        Parallel and Batch Loader for ApertureDB
    """

    def __init__(self, db, dry_run=False):

        self.db = db
        self.dry_run = dry_run

        self.type = "element"

        # Default Values
        self.batchsize  = 1
        self.numthreads = 1
        self.n_retries  = 1

        self.total_elements = 0
        self.times_arr = []
        self.ingestion_time = 0
        self.error_counter  = 0

        self.loader_filename = "loader_progress.log"

    def do_batch(self, db, data):

        q, blobs = self.generate_batch(data)

        if not self.dry_run:
            r,b = db.query(q, blobs, n_retries=self.n_retries)
            if not db.last_query_ok():
                self.error_counter += 1
            query_time = db.get_last_query_time()
        else:
            query_time = 1

        # append is thread-safe
        self.times_arr.append(query_time)

    def worker(self, thid, generator, start, end):

        db = self.db.create_new_connection()

        if thid == 0 and self.stats:
            pb = ProgressBar.ProgressBar(self.loader_filename)

        total_batches = (end - start) // self.batchsize

        if (end - start) % self.batchsize > 0:
            total_batches += 1

        for i in range(total_batches):

            batch_start = start + i * self.batchsize
            batch_end   = min(batch_start + self.batchsize, end)

            try:
                # This can be done with slices instead of list comprehension.
                # but required support from generator.
                data_for_query = [ generator[idx]
                                   for idx in range(batch_start, batch_end) ]
                self.do_batch(db, data_for_query)
            except:
                self.error_counter += 1

            if thid == 0 and self.stats:
                pb.update((i + 1) / total_batches)

    def ingest(self, generator, batchsize=1, numthreads=1, stats=False):

        self.times_arr  = []
        self.batchsize  = batchsize
        self.numthreads = numthreads
        self.stats      = stats
        self.total_elements = len(generator)

        start_time = time.time()

        elements_per_thread = math.ceil(self.total_elements / self.numthreads)

        thread_arr = []
        for i in range(self.numthreads):
            idx_start = i * elements_per_thread
            idx_end   = min(idx_start + elements_per_thread, self.total_elements)

            thread_add = Thread(target=self.worker,
                                args=(i, generator, idx_start, idx_end))
            thread_arr.append(thread_add)


        a = [ th.start() for th in thread_arr]
        a = [ th.join()  for th in thread_arr]

        self.ingestion_time = time.time() - start_time

        if self.stats:
            self.print_stats()

    def print_stats(self):

        times = np.array(self.times_arr)
        total_queries_exec = len(times)
        inserted_elements  = self.total_elements

        print("============ ApertureDB Loader Stats ============")
        print("Total time (s):", self.ingestion_time)
        print("Total queries executed:", total_queries_exec)
        print("Avg Query time (s):", np.mean(times))
        print("Query time std:", np.std (times))
        print("Avg Query Throughput (q/s)):",
                1 / np.mean(times) * self.numthreads)

        msg = "(" + self.type + "/s):"
        print("Overall insertion throughput", msg,
                self.total_elements / self.ingestion_time)

        if self.error_counter > 0:
            print("Total errors encountered:", self.error_counter)
            inserted_elements -= self.error_counter * self.batchsize
            print("Errors (%):", 100 * self.error_counter / total_queries_exec)

        print("Total inserted elements:", inserted_elements)
        print("=================================================")
